// This file is part of StarClassifier. Please see StarClassifier.java or 
// http://starclassifier.udd.be/ for license information. 
// Copyright (C) 2008 Niklas Udd

// This file loads a decision tree from Weka's plain text format. Each line in such
// a file has the following form:
//      attribute operator value[: class (totalExamples[/incorrectExamples])]
// The indentation signals the level.
// This class turned out to be more complex than intended and it is a good idea to 
// rewrite it.

package be.udd.starclassifier;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Vector;

public class DecisionTreeLoader {
	public static DecisionTree load(String filename, String defaultClass) throws FileNotFoundException, IOException {
		FileReader reader = new FileReader(filename);
		BufferedReader bufferedReader = new BufferedReader(reader);

		// Read the contents of the file and transform it into NodeLineRepresentations
		Vector<NodeLineRepresentation> nodes = new Vector<NodeLineRepresentation>();
		String line;
		while((line = bufferedReader.readLine()) != null) {
			// Find the level
			int level = 0;
			while(line.charAt(0) == '|') {
				level++;
				line = line.substring(1).trim();
			}

			// Find the attribute
			int posOfFirstSpace = line.indexOf(' ');
			String attribute = line.substring(0, posOfFirstSpace);
			
			// Find the operator
			int posOfSecondSpace = line.indexOf(' ', posOfFirstSpace + 1);
			String operator;		
			if(posOfSecondSpace == -1) {
				// This is a hack needed in order to handle the empty string as a
				// value (e.g., "language =")
				operator = line.substring(posOfFirstSpace + 1, line.length());
				posOfSecondSpace = line.length() - 1;
			} else {
				operator = line.substring(posOfFirstSpace + 1, posOfSecondSpace);
			}
			
			// Find the value
			int posOfColon = line.indexOf(':', posOfSecondSpace + 1);
			int posOfEndOfValue = line.length();
			if(posOfColon != -1) {
				posOfEndOfValue = posOfColon;
			}
			String value = line.substring(posOfSecondSpace + 1, posOfEndOfValue);
			
			
			// Find the leaf (if present)
			boolean hasLeaf = false;
			String predictedClass = null;
			double total = 0;
			double error = 0;
			if(posOfColon != -1) {
				// Find the class (in most cases it would be correct to search for the
				// third space, but this does not always work since a class name can
				// contain spaces)
				hasLeaf = true;	
				int posOfLastSpace = line.lastIndexOf(' ');
				predictedClass = line.substring(posOfColon + 2, posOfLastSpace);
				
				// Find class distribution
				int posOfSlash = line.indexOf('/', posOfLastSpace + 2);
				int posOfParenthesis = line.indexOf(')', posOfLastSpace + 2);
				if(posOfSlash != -1) {
					error = Double.parseDouble(line.substring(posOfSlash + 1, posOfParenthesis));
					posOfParenthesis = posOfSlash;
				}
				total = Double.parseDouble(line.substring(posOfLastSpace + 2, posOfParenthesis));
			}
			
			// Create a NodeLineRepresentation 
			NodeLineRepresentation currentNode;
			if(operator.equals("="))
				currentNode = new NodeLineRepresentation(level, attribute, value);
			else
				currentNode = new NodeLineRepresentation(level, attribute, operator, Integer.parseInt(value));
			if(hasLeaf)
				currentNode.setLeaf(predictedClass, total, error);
			
			nodes.add(currentNode);
		}
		
		bufferedReader.close();
		reader.close();
				
		return nodeLinesToTree(nodes, defaultClass);
	}
	
	private static DecisionTree nodeLinesToTree(Vector<NodeLineRepresentation> nodes, String defaultClass) {
		// Transform a Vector of NodeLineRepresentations to a DecisionTree  
		
		NodeLineRepresentation currentNode = nodes.get(0);
		int currentLevel = currentNode.getLevel();
		
		// Create the node
		DecisionNode currentDecisionNode;
		if(currentNode.getType() == NodeLineRepresentation.NUMERIC_NODE)
			currentDecisionNode = new DecisionNodeNumeric(currentNode.getAttribute(), currentNode.getSplitValue());
		else
			currentDecisionNode = new DecisionNodeCategoric(currentNode.getAttribute());

		// Add the childs
		String edge = null;	
		Vector<NodeLineRepresentation> subNodes = new Vector<NodeLineRepresentation>();
		for(int i = 0; i < nodes.size(); i++) {
			currentNode = nodes.get(i);
	
			if(currentNode.getLevel() <= currentLevel) {
				if(subNodes.size() > 0) {
					DecisionTree subTree = nodeLinesToTree(subNodes, defaultClass);
					currentDecisionNode.addChild(edge, subTree);
					subNodes.clear();
				}			
				if(currentNode.getLevel() < currentLevel) {
					break;
				}
			}

			if(currentNode.getLevel() == currentLevel) {
				if(currentNode.getType() == NodeLineRepresentation.NUMERIC_NODE)
					edge = currentNode.getOperator();
				else
					edge = currentNode.getValue();			
				
				if(currentNode.hasLeaf()) {
					DecisionLeaf leaf = new DecisionLeaf(currentNode.getClassName(), (currentNode.getClassName().equals(defaultClass)), currentNode.getTotal(), currentNode.getError());
					currentDecisionNode.addChild(edge, leaf);
				}
			} else if(currentNode.getLevel() > currentLevel) {
				subNodes.add(currentNode);
			}			
		}
		
		if(subNodes.size() > 0) {
			DecisionTree subTree = nodeLinesToTree(subNodes, defaultClass);
			currentDecisionNode.addChild(edge, subTree);
			subNodes.clear();
		}
		
		return currentDecisionNode;
	}
}